// Copyright John Main: TronicsBench.com
// Free for use in non-commercial projects.
#include <Servo.h>
#include <ServoEasing.hpp>

// #include "u8x8-temp-sd18b20.h"
// #include <U8x8lib.h>

// U8X8_SSD1306_128X64_NONAME_HW_I2C u8x8(/* reset=*/ U8X8_PIN_NONE);

#define DEFAULT_MICROSECONDS_FOR_0_DEGREE 500
#define DEFAULT_MICROSECONDS_FOR_180_DEGREE 2500

#define NUM_SERVOS 3

ServoEasing servo[NUM_SERVOS];
int servoAngle[NUM_SERVOS];
bool servoFlag[NUM_SERVOS];    // On/off control
byte currentServo = 0;
int servoSpeed = 100;
byte direction = 0;
bool continuous = false;

//void a(void) { }

typedef int EasingProfile_t ;

typedef enum EasingProfileEnum {
    Linear,
    Quadratic,
    Cubic,
    Quartic,
    Circular,
    Sine,
    Back,
    Elastic,
    Bounce,
    Precision,
    INVALID_PROFILE
} EasingProfileEnum_t;


#define INOUT 2
int suffix = INOUT;
EasingProfileEnum_t newProfile;

//EasingProfileEnum_t profile0 = Linear;
//EasingProfileEnum_t profile1 = Quadratic;


void setup() {
  // Initialize the serial port and servos
  Serial.begin(115200);

  Serial.println("Servo Easing Testing - Multi servos.");
  showHelp();

  servo[0].attach( 9, DEFAULT_MICROSECONDS_FOR_0_DEGREE, DEFAULT_MICROSECONDS_FOR_180_DEGREE, -90, 90);
  servo[1].attach(10, DEFAULT_MICROSECONDS_FOR_0_DEGREE, DEFAULT_MICROSECONDS_FOR_180_DEGREE, -90, 90);
  servo[2].attach(11, DEFAULT_MICROSECONDS_FOR_0_DEGREE, DEFAULT_MICROSECONDS_FOR_180_DEGREE, -90, 90);

  for (int i=0;i<NUM_SERVOS;i++) {
    servoAngle[i] = 60; // Default value.
    servoFlag[i]=true;
  }
}

void printSpeed(void) {
  Serial.print("Speed ");
  Serial.println(servoSpeed);
}

char serBuf[30];
// returns 1 if \n found, otherwise 0.
int serialCommandLoop() {
static int i = 0;

  while (Serial.available()) {
    char c = Serial.read();

    if (c != '\n') {
      serBuf[i++] = c;
      if (i == sizeof(serBuf) - 1) {
        serBuf[i] = '\0';
        break;
      }
    } else {
      serBuf[i] = '\0';
      Serial.println(serBuf);
      i = 0;
      return 1;
    }

  }
  return 0;
}


void startServos(void) {

  setSpeedForAllServos(servoSpeed);

  if (direction==0) {
    for (int i=0; i<NUM_SERVOS; i++ )
    	if(servoFlag[i]) servo[i].setEaseTo(servoAngle[i]);
  } else {
    for (int i=0; i<NUM_SERVOS; i++ )
      if(servoFlag[i]) servo[i].setEaseTo(-servoAngle[i]);
  }

  synchronizeAllServosStartAndWaitForAllServosToStop();

  direction = !direction;
}

void printServoLetter(int s) {
  Serial.print( s>1 ? "C" : (s==1 ? "B" : "A") );
}


// With newline
void printServoLetterln(int s) {
  printServoLetter(s);
  Serial.println();
}

void showAllServoData(void) {
  Serial.print("currentServo ");
  printServoLetterln(currentServo);

  Serial.print("servoSpeed ");
  Serial.println(servoSpeed);
  printProfile(newProfile);
  Serial.print(" ");
  printSuffix(suffix);
  Serial.println();

  for (int i=0; i<NUM_SERVOS; i++ ) {
    Serial.print("Servo ");
    printServoLetter(i);
    Serial.print(" Easing type ");
    printDefinitionText( servo[i].getEasingType() );
    Serial.print(" is ");
    Serial.print( servoFlag[i] ? "On" : "Off" );
    Serial.print(" has angle ");
    Serial.print( servoAngle[i] );
    Serial.println();
  }
}

void showProfileHelp(void ) {
  Serial.println("Easing profile:");
  Serial.println("1. Linear");
  Serial.println("2. Quadratic");
  Serial.println("3. Cubic");
  Serial.println("4. Quartic");
  Serial.println("5. Sine");
  Serial.println("6. Circular");
  Serial.println("7. Back");
  Serial.println("8. Elastic");
  Serial.println("9. Bounce");
  Serial.println("0. Precision");
}

void showHelp(void ) {
  showProfileHelp();
  Serial.println("?: this help.");
  Serial.println("g: Go = single mode toggle default: -60 to 60 Degrees.");
  Serial.println("t - toggle continuous.");
  Serial.println("1 ~ 9 & '0': Select profile.");
  Serial.println("i: Toggle inout profile - some don't allow all (use 'd').");
  Serial.println("a,b,c : Choose servo.");
  Serial.println("s<num> set speed for all servos.");
  Serial.println("r - Reset and stop servos - set to 0 degrees (center).");
  Serial.println("o - Toggle current servo on/off");
  Serial.println("m<num> - Set current servo angle.");
  Serial.println("d - show data.");
}

void serialCommand(void) {

    // Use the incoming byte
    char firstChar =serBuf[0];

    if (isDigit(firstChar)) {  // set profile

      newProfile = selectProfile(firstChar);

      if (newProfile != INVALID_PROFILE) setServoProfile(newProfile, suffix);

      printProfile(newProfile);
      Serial.print(" for Servo ");
      printServoLetterln(currentServo);

    } else if (firstChar == 'a') { // Select servo A
      Serial.println("Selecting servo A");
      currentServo = 0;

    } else if (firstChar == 'b') { // Select servo B
      Serial.println("Selecting servo B");
      currentServo = 1;

    } else if (firstChar == 'c') { // Select servo C
      Serial.println("Selecting servo C");
      currentServo = 2;

    } else if (firstChar == 'd') { // Show data
      showAllServoData();

    } else if (firstChar == 'g') { // Go - toggle direction of movement.
      startServos();

    } else if (firstChar == 'i') { // Choose suffix modifer
        Serial.println("next In/Out suffix");
        suffix++;
        if(suffix>=4) suffix= 0;
        printSuffix(suffix);
        Serial.println();
        setServoProfile(newProfile, suffix);
        showAllServoData();

    } else if  (firstChar == 'm') { // Max angle both ends = same
      int num = atoi(&serBuf[1]);
      if (num>=10 && num<=90) servoAngle[currentServo] = num;
      Serial.println(F("Servo angle "));
      printServoLetter(currentServo);
      Serial.println(servoAngle[currentServo]);

    } else if  (firstChar == 'o') { //Toggle a servo on/off

      servoFlag[currentServo] = !servoFlag[currentServo];

      if(servoFlag[currentServo]==false)
        servo[currentServo].setEaseTo(0); // park at 0 (center).

      printServoLetter(currentServo);
      Serial.print(" is ");
      Serial.println( servoFlag[currentServo] ? "On" : "Off" );

    } else if  (firstChar == 'r') { // Reset to center 0 Degrees and stop toggle

      for (int i=0; i<NUM_SERVOS; i++ ) {
        servo[i].setEaseTo(0);
        continuous = 0;
      }

      synchronizeAllServosStartAndWaitForAllServosToStop();

    } else if  (firstChar == 's') { // Speed

       if( isdigit(serBuf[1]) ) {
         int num = atoi(&serBuf[1]);
         if (num!=0 && num>30) servoSpeed=num;
         else Serial.println(F("Too slow - allowed but of no use!"));
         printSpeed();

       }
    } else if  (firstChar == 't') { // Auto toggle

       continuous = !continuous;

    } else if (firstChar == '?')
      showHelp();
}

static long tw = millis(); // Period before change servo direction.

void loop() {

   if ( serialCommandLoop() ) serialCommand();

   if (continuous && millis() - tw>1000) {
     tw = millis();
     startServos();
   }
}





// Helper function to select an easing profile
// if not a number input is ignored and returns INVALID_PROFILE as indicator.
EasingProfileEnum_t selectProfile(char incomingByte) {

  // Set the easing profile based on the incoming byte
  if (incomingByte == '1') {
    return Linear;
  } else if (incomingByte == '2') {
    return Quadratic;
  } else if (incomingByte == '3') {
    return Cubic;
  } else if (incomingByte == '4') {
    return Quartic;
  } else if (incomingByte == '5') {
    return Sine;
  } else if (incomingByte == '6') {
    return Circular;
  } else if (incomingByte == '7') {
    return Back;
  } else if (incomingByte == '8') {
    return Elastic;
  } else if (incomingByte == '9') {
    return Bounce;
  } else if (incomingByte == '0') {
    return Precision;
  } return INVALID_PROFILE;
}


void setServoProfile(EasingProfileEnum_t newProfile, int suffix) {

  switch(suffix) {

    case 0: setServoProfile_IN(newProfile);
      break;

    case 1:  setServoProfile_OUT(newProfile);
      break;

    case 2: setServoProfile_IN_OUT(newProfile);
      break;

    case 3:  setServoProfile_BOUNCING(newProfile);
      break;
  }
}


void setServoProfile_IN_OUT(EasingProfileEnum_t newProfile) {
  switch (newProfile) {
    case Linear:
      servo[currentServo].setEasingType(EASE_LINEAR); // Is the same for IN_OUT
      break;

    case Quadratic:
      servo[currentServo].setEasingType(EASE_QUADRATIC_IN_OUT);
      break;

    case Cubic:
      servo[currentServo].setEasingType(EASE_CUBIC_IN_OUT);
      break;

    case Quartic:
      servo[currentServo].setEasingType(EASE_QUARTIC_IN_OUT);
      break;

    case Circular:
      servo[currentServo].setEasingType(EASE_CIRCULAR_IN_OUT);
      break;

    case Sine:
      servo[currentServo].setEasingType(EASE_SINE_IN_OUT);
      break;

    case Back:
      servo[currentServo].setEasingType(EASE_BACK_IN_OUT);
      break;

    case Elastic:
      servo[currentServo].setEasingType(EASE_ELASTIC_IN_OUT);
      break;

    case Bounce:
      servo[currentServo].setEasingType(EASE_BOUNCE_OUT); // No in prob as you want at end
      break;

    case Precision:
      servo[currentServo].setEasingType(EASE_PRECISION_OUT); // Meaining of precision?
      break;

    default:
      printf("Invalid easing profile");
      break;
  } // switch
}

void setServoProfile_IN(EasingProfileEnum_t newProfile) {
  switch (newProfile) {
    case Linear:
      servo[currentServo].setEasingType(EASE_LINEAR); // Is the same for IN_OUT
      break;

    case Quadratic:
      servo[currentServo].setEasingType(EASE_QUADRATIC_IN);
      break;

    case Cubic:
      servo[currentServo].setEasingType(EASE_CUBIC_IN);
      break;

    case Quartic:
      servo[currentServo].setEasingType(EASE_QUARTIC_IN);
      break;

    case Circular:
      servo[currentServo].setEasingType(EASE_CIRCULAR_IN);
      break;

    case Sine:
      servo[currentServo].setEasingType(EASE_SINE_IN);
      break;

    case Back:
      servo[currentServo].setEasingType(EASE_BACK_IN);
      break;

    case Elastic:
      servo[currentServo].setEasingType(EASE_ELASTIC_IN);
      break;

    case Bounce:
      servo[currentServo].setEasingType(EASE_BOUNCE_OUT); // No in prob as you want at end
      break;

    case Precision:
      servo[currentServo].setEasingType(EASE_PRECISION_OUT); // Meaining of precision?
      break;

    default:
      printf("Invalid easing profile");
      break;
  } // switch
}

void setServoProfile_OUT(EasingProfileEnum_t newProfile) {
  switch (newProfile) {
    case Linear:
      servo[currentServo].setEasingType(EASE_LINEAR); // Is the same for IN_OUT
      break;

    case Quadratic:
      servo[currentServo].setEasingType(EASE_QUADRATIC_OUT);
      break;

    case Cubic:
      servo[currentServo].setEasingType(EASE_CUBIC_OUT);
      break;

    case Quartic:
      servo[currentServo].setEasingType(EASE_QUARTIC_OUT);
      break;

    case Circular:
      servo[currentServo].setEasingType(EASE_CIRCULAR_OUT);
      break;

    case Sine:
      servo[currentServo].setEasingType(EASE_SINE_OUT);
      break;

    case Back:
      servo[currentServo].setEasingType(EASE_BACK_OUT);
      break;

    case Elastic:
      servo[currentServo].setEasingType(EASE_ELASTIC_OUT);
      break;

    case Bounce:
      servo[currentServo].setEasingType(EASE_BOUNCE_OUT); // No in prob as you want at end
      break;

    case Precision:
      servo[currentServo].setEasingType(EASE_PRECISION_OUT); // Meaining of precision?
      break;

    default:
      printf("Invalid easing profile");
      break;
  } // switch
}

void setServoProfile_BOUNCING(EasingProfileEnum_t newProfile) {
  switch (newProfile) {
    case Linear:
      servo[currentServo].setEasingType(EASE_LINEAR); // Is the same for _IN_OUT
      break;

    case Quadratic:
      servo[currentServo].setEasingType(EASE_QUADRATIC_BOUNCING);
      break;

    case Cubic:
      servo[currentServo].setEasingType(EASE_CUBIC_BOUNCING);
      break;

    case Quartic:
      servo[currentServo].setEasingType(EASE_QUARTIC_BOUNCING);
      break;

    case Circular:
      servo[currentServo].setEasingType(EASE_CIRCULAR_BOUNCING);
      break;

    case Sine:
      servo[currentServo].setEasingType(EASE_SINE_BOUNCING);
      break;

    case Back:
      servo[currentServo].setEasingType(EASE_BACK_BOUNCING);
      break;

    case Elastic:
      servo[currentServo].setEasingType(EASE_ELASTIC_BOUNCING);
      break;

    case Bounce:
      servo[currentServo].setEasingType(EASE_BOUNCE_OUT); // Replace _IN_OUT with _BOUNCING
      break;

    case Precision:
      servo[currentServo].setEasingType(EASE_PRECISION_OUT); // Meaning of precision?
      break;

    default:
      printf("Invalid easing profile");
      break;
  } // switch
}

void printProfile (EasingProfileEnum_t profile) {
   switch (profile) {
        case Linear:
            Serial.print("Linear");
            break;
        case Quadratic:
            Serial.print("Quadratic");
            break;
        case Cubic:
            Serial.print("Cubic");
            break;
        case Quartic:
            Serial.print("Quartic");
            break;
        case Circular:
            Serial.print("Circular");
            break;
        case Sine:
            Serial.print("Sine");
            break;
        case Back:
            Serial.println("Back");
            break;
        case Elastic:
            Serial.print("Elastic");
            break;
        case Bounce:
            Serial.print("Bounce");
            break;
        case Precision:
            Serial.print("Precision");
            break;
        case INVALID_PROFILE:
            Serial.print("INVALID_PROFILE");
            break;
        default:
            Serial.print("Unknown");
            break;
    }
}


void printSuffix (int suffix) {
   switch (suffix) {
        case 0:
            Serial.print("(In)");
            break;
        case 1:
            Serial.println("(OUT)");
            break;
        case 2:
            Serial.println("(IN_OUT)");
            break;
        case 3:
            Serial.println("(BOUNCING)");
            break;

        default:
            Serial.println("(Unknown)");
            break;
    }
}

// code definition returned by getEasingType back to text string.
void printDefinitionText(uint8_t value) {
  switch (value) {
//    case CALL_STYLE_DIRECT:
//    case CALL_STYLE_IN:
//      Serial.print(F("In"));
//      break;
//    case CALL_STYLE_OUT:
//      Serial.print(F("Out"));
//      break;
//    case CALL_STYLE_IN_OUT:
//      Serial.print(F("In Out"));
//      break;
//    case CALL_STYLE_BOUNCING_OUT_IN:
//      Serial.print(F("Bouncing Out In"));
//      break;
//    case CALL_STYLE_MASK:
//      Serial.print(F("CALL_STYLE_MASK"));
//      break;
//    case EASE_TYPE_MASK:
//      Serial.print(F("EASE_TYPE_MASK"));
//      break;
    case EASE_LINEAR:
      Serial.print(F("Linear"));
      break;
    #if defined(ENABLE_EASE_QUADRATIC)
      case EASE_QUADRATIC_IN:
        Serial.print(F("Quadratic (In)"));
        break;
      case EASE_QUADRATIC_OUT:
        Serial.print(F("Quadratic (Out)"));
        break;
      case EASE_QUADRATIC_IN_OUT:
        Serial.print(F("Quadratic (In Out)"));
        break;
      case EASE_QUADRATIC_BOUNCING:
        Serial.print(F("Quadratic (Bouncing)"));
        break;
    #endif
    #if defined(ENABLE_EASE_CUBIC)
      case EASE_CUBIC_IN:
        Serial.print(F("Cubic (In)"));
        break;
      case EASE_CUBIC_OUT:
        Serial.print(F("Cubic (Out)"));
        break;
      case EASE_CUBIC_IN_OUT:
        Serial.print(F("Cubic (In Out)"));
        break;
      case EASE_CUBIC_BOUNCING:
        Serial.print(F("Cubic (Bouncing)"));
        break;
    #endif
    #if defined(ENABLE_EASE_QUARTIC)
      case EASE_QUARTIC_IN:
        Serial.print(F("Quartic (In)"));
        break;
      case EASE_QUARTIC_OUT:
        Serial.print(F("Quartic (Out)"));
        break;
      case EASE_QUARTIC_IN_OUT:
        Serial.print(F("Quartic (In Out)"));
        break;
      case EASE_QUARTIC_BOUNCING:
        Serial.print(F("Quartic (Bouncing)"));
        break;
    #endif
    #if defined(ENABLE_EASE_USER)
      case EASE_USER_IN:         // same as case EASE_USER_DIRECT:
        Serial.print(F("User (In/Direct)"));
        break;
      case EASE_USER_OUT:
        Serial.print(F("User (Out)"));
        break;
      case EASE_USER_IN_OUT:
        Serial.print(F("User (In Out)"));
        break;
      case EASE_USER_BOUNCING:
        Serial.print(F("User (Bouncing)"));
        break;
    #endif
    case EASE_DUMMY_MOVE:
      Serial.print(F("Dummy Move"));
      break;
    #if defined(ENABLE_EASE_SINE)
      case EASE_SINE_IN:
        Serial.print(F("Sine (In)"));
        break;
      case EASE_SINE_OUT:
        Serial.print(F("Sine (Out)"));
        break;
      case EASE_SINE_IN_OUT:
        Serial.print(F("Sine (In Out)"));
        break;
      case EASE_SINE_BOUNCING:
        Serial.print(F("Sine (Bouncing)"));
        break;
    #endif
    #if defined(ENABLE_EASE_CIRCULAR)
      case EASE_CIRCULAR_IN:
        Serial.print(F("Circular (In)" ));
        break;
      case EASE_CIRCULAR_OUT:
        Serial.print(F("Circular (Out)"));
        break;
      case EASE_CIRCULAR_IN_OUT:
        Serial.print(F("Circular (In Out)"));
        break;
      case EASE_CIRCULAR_BOUNCING:
        Serial.print(F("Circular (Bouncing)"));
        break;
    #endif
    #if defined(ENABLE_EASE_BACK)
      case EASE_BACK_IN:
        Serial.print(F("Back (In)" ));
        break;
      case EASE_BACK_OUT:
        Serial.print(F("Back (Out)"));
        break;
      case EASE_BACK_IN_OUT:
        Serial.print(F("Back (In Out)"));
        break;
      case EASE_BACK_BOUNCING:
        Serial.print(F("Back (Bouncing)"));
        break;
    #endif
    #if defined(ENABLE_EASE_ELASTIC)
      case EASE_ELASTIC_IN:
        Serial.print(F("Elastic (In)" ));
        break;
      case EASE_ELASTIC_OUT:
        Serial.print(F("Elastic (Out)"));
        break;
      case EASE_ELASTIC_IN_OUT:
        Serial.print(F("Elastic (In Out)"));
        break;
      case EASE_ELASTIC_BOUNCING:
        Serial.print(F("Elastic (Bouncin)g"));
        break;
    #endif

    #if defined(ENABLE_EASE_BOUNCE)
      case EASE_BOUNCE_IN:
        Serial.print(F("EASE_BOUNCE (In)"));
        break;
      case EASE_BOUNCE_OUT:
        Serial.print(F("EASE_BOUNCE (Out)"));
        break;
     #endif

    #if defined(ENABLE_EASE_PRECISION)
       case EASE_PRECISION_IN:
        Serial.print(F("EASE_PRECISION_(In)"));
        break;
      case EASE_PRECISION_OUT:
        Serial.print(F("EASE_PRECISION_(Out)"));
        break;
    #endif
    }
}